#include <fstream.h>
#include <windows.h>
#include <gl\gl.h>
#include <gl\glu.h>
#include <gl\glaux.h>
#include <math.h>

#define ROOM_WIDTH  8
#define ROOM_HEIGHT 2
#define SPHERE_SIZE 1
#define BORDER	     0.5
#define STEP		  0.2
#define A_STEP		  4

UINT timer;
HINSTANCE MainInstance;
GLfloat x = -ROOM_WIDTH + BORDER, y = 0.0, angle = 30.0;
unsigned char *tex1, *tex2;

const GLfloat no_mat[] = { 0.0, 0.0, 0.0, 1.0 };
const GLfloat mat_grey[] = { 0.2, 0.2, 0.2, 0.7 };
const GLfloat mat_white[] = { 1.0, 1.0, 1.0, 1.0 };
const GLfloat mat_blue[] = { 0.2, 0.2, 1.0, 0.7 };
const GLfloat mat_red[] = { 1.0, 0.2, 0.2, 0.7 };
const GLfloat mat_yellow[] = { 1.0, 1.0, 0.2, 0.7 };
const GLfloat no_shininess[] = { 0.0 };
const GLfloat mat_specular[] = { 0.7, 0.7, 0.7, 1.0 };
const GLfloat mat_ambient[] = { 0.5, 0.5, 0.5, 1.0 };
const GLfloat low_shininess[] = { 2.0 };
const GLfloat RAD = M_PI / 180;

void setDCPixelFormat(HDC hdc)
{
  static PIXELFORMATDESCRIPTOR pfd =
  {
    sizeof(PIXELFORMATDESCRIPTOR),          // sizeof this structure
    1,                                      // version number
    PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL
    | PFD_DOUBLEBUFFER,                     // flags
    PFD_TYPE_RGBA,                          // RGBA pixel values
    16,                                     // 24-bit color
    0, 0, 0, 0, 0, 0,                       // don't care about these
    0, 0,                                   // no alpha buffer
    0, 0, 0, 0, 0,                          // no accumulation buffer
    16,                                     // 32-bit depth buffer
    0,                                      // no stencil buffer
    0,                                      // no auxiliary buffers
    PFD_MAIN_PLANE,                         // layer type
    0,                                      // reserved(must be 0)
    0, 0, 0                                 // no layer masks
  };
  int pixelFormat;

  pixelFormat = ChoosePixelFormat(hdc, &pfd);
  SetPixelFormat(hdc, pixelFormat, &pfd);
  DescribePixelFormat(hdc, pixelFormat, sizeof(PIXELFORMATDESCRIPTOR), &pfd);
}

void InitializeGL()
{
  GLfloat ambient[] = { 0.0, 0.0, 0.0, 1.0 };
  GLfloat diffuse[] = { 1.0, 1.0, 1.0, 1.0 };
  GLfloat position[] = { 0.0, 0.0, 5.0, 0.0 };
  GLfloat lmodel_ambient[] = { 0.4, 0.4, 0.4, 1.0 };
  GLfloat local_view[] = { 0.0 };

  glFrontFace(GL_CW);
  glCullFace(GL_BACK);
  glEnable(GL_CULL_FACE);

//  glDepthFunc(GL_LEQUAL);
//  glEnable(GL_DEPTH_TEST);
  glClearColor(0.7, 0.7, 1.0, 0.0);

  glLightfv(GL_LIGHT0, GL_AMBIENT, ambient);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuse);
  glLightfv(GL_LIGHT0, GL_POSITION, position);
  glLightModelfv(GL_LIGHT_MODEL_AMBIENT, lmodel_ambient);
  glLightModelfv(GL_LIGHT_MODEL_LOCAL_VIEWER, local_view);

  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);
  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_FASTEST);
}

void DrawRoom()
{
  glEnable(GL_TEXTURE_2D);

  glTexImage2D(GL_TEXTURE_2D, 0, 3, 256, 256, 0, GL_RGB, GL_UNSIGNED_BYTE, tex1);
  glPushMatrix();
  glMaterialfv(GL_FRONT, GL_AMBIENT, mat_ambient);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, mat_white);
  glMaterialfv(GL_FRONT, GL_SPECULAR, no_mat);
  glMaterialfv(GL_FRONT, GL_SHININESS, no_shininess);
  glMaterialfv(GL_FRONT, GL_EMISSION, no_mat);

  for (int i = 0; i < 4; i++)
  {
    glBegin(GL_QUADS);
      glNormal3f(0, 0, 1);
      glTexCoord2f(0, ROOM_HEIGHT / 2);
      glVertex3f(-ROOM_WIDTH, ROOM_HEIGHT, -ROOM_WIDTH);
      glTexCoord2f(ROOM_WIDTH, ROOM_HEIGHT / 2);
      glVertex3f(ROOM_WIDTH, ROOM_HEIGHT, -ROOM_WIDTH);
      glTexCoord1f(ROOM_WIDTH);
      glVertex3f(ROOM_WIDTH, 0, -ROOM_WIDTH);
      glTexCoord1f(0);
      glVertex3f(-ROOM_WIDTH, 0, -ROOM_WIDTH);
    glEnd();
    glRotatef(90, 0, 1, 0);
  }
  glPopMatrix();

  glTexImage2D(GL_TEXTURE_2D, 0, 3, 256, 256, 0, GL_RGB, GL_UNSIGNED_BYTE, tex2);
  glPushMatrix();
  glRotatef(-90, 1, 0, 0);
  glDisable(GL_LIGHTING);
  glBegin(GL_QUADS);
    glNormal3f(0, 0, 1);
    glTexCoord2f(0, ROOM_WIDTH);
    glVertex2f(-ROOM_WIDTH, ROOM_WIDTH);
    glTexCoord2f(ROOM_WIDTH, ROOM_WIDTH);
    glVertex2f(ROOM_WIDTH, ROOM_WIDTH);
    glTexCoord1f(ROOM_WIDTH);
    glVertex2f(ROOM_WIDTH, -ROOM_WIDTH);
    glTexCoord1f(0);
    glVertex2f(-ROOM_WIDTH, -ROOM_WIDTH);
  glEnd();
  glEnable(GL_LIGHTING);
  glPopMatrix();

  glDisable(GL_TEXTURE_2D);
}

void DrawSphere(GLfloat r, int slices, int segments)
{
  GLfloat old_x = r, old_y = 0, x, y;
  GLfloat C1 = 360.0 / (GLfloat)segments, C2 = 2.0 * r / (GLfloat)slices;
  GLfloat old_z = -r, z = -r, r2, old_r = 0;
  int i, j = -slices / 2;

    for (; j < slices / 2; j++)
    {
      z += C2;
      if (r * r - z * z < 0)
        r2 = 0;
      else
        r2 = sqrt(r * r - z * z);
      for (i = 0; i < segments + 1; i++)
      {
        x = cos(i * C1 * RAD);
        y = sin(i * C1 * RAD);
        glBegin(GL_QUADS);
          glNormal3f(-x, -y, -z);
          glVertex3f(x * r2, y * r2, z);
          glNormal3f(-old_x, -old_y, -z);
          glVertex3f(old_x * r2, old_y * r2, z);
          glNormal3f(-old_x, -old_y, -old_z);
          glVertex3f(old_x * old_r, old_y * old_r, old_z);
          glNormal3f(-x, -y, -old_z);
          glVertex3f(x * old_r, y * old_r, old_z);
        glEnd();
        old_x = x;
        old_y = y;
      }
      old_z = z;
      old_r = r2;
    }
}

void DrawSpheres()
{
  glMaterialfv(GL_FRONT, GL_AMBIENT, mat_ambient);
  glMaterialfv(GL_FRONT, GL_SPECULAR, mat_specular);
  glMaterialfv(GL_FRONT, GL_EMISSION, no_mat);
  glMaterialfv(GL_FRONT, GL_SHININESS, low_shininess);

  glPushMatrix();

  glTranslatef(ROOM_WIDTH / 2, SPHERE_SIZE, ROOM_WIDTH / 2);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, mat_blue);
  glCallList(1);

  glTranslatef(0, 0, -ROOM_WIDTH);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, mat_red);
  glCallList(1);

  glTranslatef(-ROOM_WIDTH, 0, 0);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, mat_yellow);
  glCallList(1);

  glTranslatef(0, 0, ROOM_WIDTH);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, mat_grey);
  glCallList(1);

  glPopMatrix();
}

void drawScene(HDC hDC)
{
  glClear(GL_COLOR_BUFFER_BIT/* | GL_DEPTH_BUFFER_BIT*/);
  glMatrixMode(GL_MODELVIEW);

  glLoadIdentity();
  glRotatef(angle, 0, 1, 0);
  glTranslatef(y, -0.5, x);

  DrawRoom();
  DrawSpheres();

  SwapBuffers(hDC);
  glFlush();
}

LONG WINAPI wndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
  static HDC hdc;
  static HGLRC hrc;
  PAINTSTRUCT ps;
  GLdouble aspect;
  GLsizei width, height;

  switch(msg)
  {
    case WM_KEYDOWN:
    	switch (wParam)
      {
        case 0x27: angle += A_STEP; break;
        case 0x25: angle -= A_STEP; break;
        case 0x26: x += cos(angle * RAD) * STEP;
        				 y -= sin(angle * RAD) * STEP;	break;
        case 0x28: x -= cos(angle * RAD) * STEP;
        				 y += sin(angle * RAD) * STEP;	break;
      }
      if (x > ROOM_WIDTH - BORDER) x = ROOM_WIDTH - BORDER;
      if (x < -ROOM_WIDTH + BORDER) x = -ROOM_WIDTH + BORDER;
      if (y > ROOM_WIDTH - BORDER) y = ROOM_WIDTH - BORDER;
      if (y < -ROOM_WIDTH + BORDER) y = -ROOM_WIDTH + BORDER;

      return 0;

    case WM_CREATE:
      hdc = GetDC(hWnd);
      setDCPixelFormat(hdc);
      hrc = wglCreateContext(hdc);
      wglMakeCurrent(hdc, hrc);
      InitializeGL();
      glNewList(1, GL_COMPILE);
      DrawSphere(SPHERE_SIZE, 20, 20);
      glEndList();
      timer = SetTimer(hWnd, 1, 20, NULL);
      return 0;

    case WM_SIZE:
      width = (GLsizei)LOWORD(lParam);
      height = (GLsizei)HIWORD(lParam);
      if (!width || !height) return 0;

      aspect = (GLdouble)width /(GLdouble)height;
      glMatrixMode(GL_PROJECTION);
      glLoadIdentity();
      gluPerspective(30.0, aspect, 0.0, 32.0);
      glViewport(0, 0, width, height);

      return 0;

    case WM_PAINT:
      BeginPaint(hWnd, &ps);
      drawScene(hdc);
      EndPaint(hWnd, &ps);
      return 0;

    case WM_DESTROY:
      wglMakeCurrent(NULL, NULL);
      wglDeleteContext(hrc);
      KillTimer(hWnd, timer);
      PostQuitMessage(0);
      return 0;

    case WM_TIMER:
      InvalidateRect(hWnd, NULL, FALSE);
      return 0;
  }

  return DefWindowProc(hWnd, msg, wParam, lParam);
}

int LoadTexture(unsigned char *&texture, char *filename, long size)
{
  ifstream in(filename, ios::binary);
  if (!in) return 0;
  in.seekg(54, ios::beg);
  texture = new unsigned char[size];
  in.read(texture, size);
  in.close();
  return 1;
}

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR cmdLine, int cmdShow)
{
  static char appName [] = "Spheres - by HackMaster";
  WNDCLASS wc;
  HWND hWnd;
  MSG msg;

  if (!LoadTexture(tex1, "bricks.bmp", 196608))
    return 1;
  if (!LoadTexture(tex2, "asphalt.bmp", 196608))
    return 1;

  MainInstance = hInstance;

  wc.style = CS_HREDRAW | CS_VREDRAW;
  wc.lpfnWndProc =(WNDPROC)wndProc;
  wc.cbClsExtra = 0;
  wc.cbWndExtra = 0;
  wc.hInstance = hInstance;
  wc.hIcon = LoadIcon(NULL, IDI_APPLICATION);
  wc.hCursor = LoadCursor(NULL, IDC_ARROW);
  wc.hbrBackground =(HBRUSH)(COLOR_WINDOW + 1);
  wc.lpszMenuName = (LPCTSTR)0;
  wc.lpszClassName = appName;

  RegisterClass(&wc);

  hWnd = CreateWindow(appName, appName,
                      WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN | WS_CLIPSIBLINGS,
                      CW_USEDEFAULT, CW_USEDEFAULT, 800,
                      600, HWND_DESKTOP, NULL, hInstance, NULL);

  ShowWindow(hWnd, cmdShow);
  UpdateWindow(hWnd);

  while(GetMessage(&msg, NULL, 0, 0))
  {
    TranslateMessage(&msg);
    DispatchMessage(&msg);
  }

  delete tex1;
  delete tex2;

  return msg.wParam;
}
